#!/usr/bin/perl
#  File: HTTPtest.pl
#
#  Author: G. Wade Johnson
#
#  Copyright 2001 by Telescan, Inc.
#  Released under the same license as Perl.
#  
#  Warning: This program is under construction.
#  It may be changed at any time.

use FindBin;
use lib $FindBin::Bin;
use LWP::UserAgent;
use HTTP::Cookies;
use XML::Parser;
use XMLdata;
use URI::Escape;
use Getopt::Std;
use File::Spec;

$VERSION = 0.782;

use constant STOP_EXCEPTION => "msg-stop\n";
use constant BOUNDARY       => '1234httptest-boundary-gwj';

my $CompileMsg = '';
BEGIN
 {
  $^W = 1;

  if($^O eq 'MSWin32')
   {
    use File::DosGlob 'glob';
    my @files;
    @ARGV = map { (/[\*\?]/ && (@files = glob $_)) ? @files : $_ } @ARGV
                                                                if $] > 5.00307;
   }

  #-----------------------------------------------------------------
  # Attempt to 'use' the Time::HiRes module. If it is not available,
  # we continue. The module is not necessary, it just gives better
  # resolution if we have it.
  eval { require Time::HiRes; Time::HiRes->import( 'time' ); };
  $CompileMsg .= "HiRes Timing not available.\n" if $@;
 }

use strict;
use vars qw/%opts %Options $VERSION/;

getopts( 'vwqVlo:I:f:jJ:', \%opts ) || usage();

if($opts{V})
 {
  print "$0: Version $VERSION\n";
  exit( 0 );
 }

my $ResponseTime  = 0;
my $DefaultLibDir = "$FindBin::Bin/lib";
my %Params        = ();
my %CodeLibIDs    = ();
my %SubTests      = ();
my %CmdLine       = ();
my $ScriptStart   = time;
my @Counters      = (qw/ok skip tests curr/, 'not ok');

extract_cmd_line();
expand_list_file();

usage() unless @ARGV;

print STDERR $CompileMsg if $CompileMsg and 'verbose' eq option( 'msgs' );

my $parser = new XML::Parser( Style=>'Objects', Pkg=>'tst' );
XMLdata::set_package( 'tst' );

my $ua = new LWP::UserAgent;
die "Unable to create user agent.\n" unless $ua;

my @TopLevelElements = qw/battery testset param test test-set session codelib
                          include inclusion option subtest-def/;

#-----------------------------------------------
# Define handling of top-level elements
my %TSElement =
 (
  param         => sub { },  # already handled
  test          => sub { run_test( $ua, @_ ); },
  session       => \&execute_a_test_set,
  'test-set'    => \&execute_a_test_set,
  codelib       => \&compile_lib,
  'subtest-def' => \&build_subtest_definition,
  include       => sub { },  # already handled
  option        => sub { die "An option element found out of place.\n"; }
 );
XMLdata::DeclareElements( @TopLevelElements );

my @TestComponents = qw/request response body auth not or and or-all and-all
                        status header content time code msg filter call skip/;

#-----------------------------------------------
# Define handling of subtest elements
my %Tests =
 (  # general
  'not'	    => \&validate_not,
  'or'	    => \&validate_or,
  'and'	    => \&validate_and,
  'or-all'  => \&validate_or_all,
  'and-all' => \&validate_and_all,
  'call'    => \&call_subtest_definition,
    # protocol-specific
  status    => \&validate_status,
  header    => \&validate_header,
  content   => \&validate_content,
  'time'    => \&validate_time,
  code      => \&validate_code,
  msg       => \&send_msg,
  filter    => \&execute_filter,
 );
XMLdata::DeclareElements( @TestComponents );

my %TotCounts = init_counts( sets => 0 );

#-----------------------------------------------
# Main file processing loop.
#  invariant: @ARGV contains battery filenames to process.
while(@ARGV)
 {
  my $testfile = shift;
  my $doc = $parser->parsefile( $testfile ) or die "Unable to read battery file \'$testfile\'.\n";
  my $battery = XMLdata::root( $doc ) or die "Unable to read battery file \'$testfile\'.\n";

  # Extract options
  extract_options( $battery );
  add_command_line_overrides();
  open_logfile();

  # Expand includes at this point.
  expand_includes( $battery );
  my $err = invalid_battery( $battery );
  die "Battery \"$testfile\" has errors: $err\n" if $err;

  # initialize per-file data
  params_initialize();
  %SubTests = ();

  my %Counts = init_counts( tests => count_tests( $battery ) );

  printlog( '# ', $battery->{label}, "\n" );
  printlogv( "1..$Counts{tests}\n" );

  setup_params( $battery );

  eval
   {
    update_counts( \%Counts, process_battery( $battery->get_children() ) );
   };

  if($@)
   {
    die $@ if $@ ne STOP_EXCEPTION;
    print "\n";
   }
  else
   {
    if($Counts{'not ok'})
     {
      printlog( "Failed ", ($Counts{'not ok'}), "/$Counts{tests}\n" );
     }
    if($Counts{skip})
     {
      printlog( "Skipped ", ($Counts{skip}), "/$Counts{tests}\n" );
     }
   }

  update_counts( \%TotCounts, \%Counts );
  ++$TotCounts{sets};
  close_logfile();
 }

my $ScriptTime = time - $ScriptStart;

option( 'logging' => 0 );  # force logging off.

if($TotCounts{ok} == $TotCounts{tests})
 {
  if(1 < $TotCounts{sets})
   {
    printlog( "All tests successful.\nFiles=$TotCounts{sets}, Tests=$TotCounts{tests}.\n" );
   }
  else
   {
    printlog( "All tests successful.\nTests=$TotCounts{tests}.\n" );
   }
 }
elsif($TotCounts{'not ok'})
 {
  printlog( "FAILED tests\n\tFailed $TotCounts{'not ok'}/$TotCounts{tests}\n" );
 }
if($TotCounts{skip})
 {
  printlog( "SKIPPED tests.\n\tSkipped $TotCounts{skip}/$TotCounts{tests}\n" );
 }
printlog( sprintf "# Elapsed time: %.2f seconds\n", $ScriptTime );


#--------------------------------------------------------
#  Subroutines
#--------------------------------------------------------

#--------------------------------------------------------
#  Update one count structure with the data contained in
#  another.
#
#  Input:  $rAcc  - the accumlator count structure
#          $rNew  - new counts to accumulate
sub  update_counts
 {
  my $rAcc = shift;
  my $rNew = shift;
  
  foreach my $cntr (@Counters)
   {
    $rAcc->{$cntr} += $rNew->{$cntr} if exists $rNew->{$cntr};
   }
 }

#--------------------------------------------------------
#  Supply initial count structure
#
#  Input:  Optional list of initial values
sub  init_counts
 {
  my %Counts = (
                (map { $_ => 0 } @Counters),
                @_
               );

  wantarray ? %Counts : \%Counts;
 }


#--------------------------------------------------------
#  Count the test elements in the supplied container
#
#  Input:  $container  - the container object to count
#
#  Output: number of test elements found in container
sub   count_tests
 {
  my $container = shift;

  return scalar( @{$container->get_descendents_by_tag( 'test' )} );
 }

#--------------------------------------------------------
#  Process several tests
#
#  Input:  A list of top-level elements to process.
#
#  Output: the number of successful tests
sub  process_battery
 {
  my $Counts = init_counts();
  process_set_of_tests( $Counts, @_ );

  $Counts;
 }


#--------------------------------------------------------
#  Process several tests
#
#  Input:   $Counts - reference to a counts structure
#           A list of top-level elements to process.
sub    process_set_of_tests
 {
  my $Counts  = shift;
  
  # invariant: @_ contains the remaining top-level test elements
  foreach my $el (@_)
   {
    next unless $el->is_element();

    my $tag = $el->tagname();

    if(exists $TSElement{$tag})
     {
      $TSElement{$tag}->( $el, $Counts );
     }
    else
     {
      die "Unrecognized element '$tag'.\n";
     }
   }
 }
  
  
#--------------------------------------------------------
# Initialize Options variables
sub   initialize_options
 {
  %Options = ( incpath => [],
               logging => 0,
               logfile => 'HTTPtest.log',
               msgs    => '',
             );
 }

#--------------------------------------------------------
#  Extract the option elements, setting the appropriate
#  internal variables
#
#  Input:  reference to the battery object
#
#  Currently supported options
#    incpath   - include path                   (;-separated list)
#    logging   - turn logging on/off            (yes|no)
#    logfile   - define file to log output
#    msgs      - specify how verbose the output (verbose|quiet)?
sub  extract_options
 {
  my $battery  = shift;
  my $children = $battery->get_children();

  initialize_options();   # reset options variables for each new file

  while(@{$children})
   {
    my $child = shift @{$children};

    next   unless $child->is_element();

    if('option' ne $child->tagname())
     {
      unshift @{$children}, $child;
      return;
     }

    # need to drive this with a data structure.
    if('incpath' eq $child->{name})
     {
      my $prev = option( $child->{name} );
      option( $child->{name} => [ $prev ? @{$prev} : (),
                                  split( ';', $child->{value} )
                                ]
	    );
     }
    elsif('logging' eq $child->{name})
     {
      option( $child->{name} => ('yes' eq $child->{value}) );
     }
    else
     {
      option( $child->{name} => $child->{value} );
     }
   }
 }


#--------------------------------------------------------
#  Override the internal options based on the command line
#  options.
sub   add_command_line_overrides
 {
  option( 'msgs' => ''        ) if defined $opts{w};
  option( 'msgs' => 'verbose' ) if defined $opts{v};
  option( 'msgs' => 'quiet'   ) if defined $opts{q};

  option( 'logging'   => defined $opts{l} || defined $opts{o} );
  option( 'logfile'   => $opts{o} ) if defined $opts{o};
  option( 'cookiejar' => $opts{J} || '' ) if exists $opts{J} or exists $opts{j};

  # Reverse needed to keep search order intact.
  my $incpath = $Options{incpath} || [];
  unshift @{$incpath}, split( ';', $opts{I} ) if defined $opts{I};
  push @{$incpath}, $DefaultLibDir;
  option( 'incpath' => $incpath );

  if(option_exists( 'cookiejar' ))
   {
    my @cjparms = ();
    @cjparms = ( file => option( 'cookiejar' ), autosave => 1 );
    my $cj = HTTP::Cookies->new( @cjparms );
    $ua->cookie_jar( $cj );
   }
 }




#--------------------------------------------------------
#  Perform one test from the battery and report on
#  results
#
#  Input:  $ua   - reference to an LWP::UserAgent object
#          $test - reference to a test object to execute
#          $Counts - reference to a count structure
sub   run_test
 {
  my $ua     = shift;
  my $test   = shift;
  my $Counts = shift;   
  my $label = $test->{label} ? " - $test->{label}" : '';

  $Counts->{curr}++;
  
  my $result = evaluate_test( $ua, $test );
  printlogv( "$result $Counts->{curr}$label\n" );

  $Counts->{$result}++;
 }


#--------------------------------------------------------
#  Perform one test from the battery
#
#  Input:  $ua   - reference to an LWP::UserAgent object
#          $test - reference to a test object to execute
#
#  Output: "ok" if successful, "not ok" if failed, or "skip"
#          if optional skip test succeeds
sub  evaluate_test
 {
  my $ua     = shift;
  my $test   = shift;

  return "skip" if handle_skips( $test->get_children_by_tag( 'skip' ) );

  my $treq   = $test->get_child_by_tag( 'request', 0 );

  my $url    = resolve_params( $treq->{url} );
  my $method = $treq->{method} || "GET";
  my $req;

  printlogvv( "\n  $method $url\n" );

  my $hdr   = new HTTP::Headers;

  my $auth  = $treq->get_child_by_tag( 'auth' );
  if($auth)
   {
    require MIME::Base64;
    my $user = resolve_params( $auth->{user} );
    my $pass = resolve_params( $auth->{password} );

    my $authval = MIME::Base64::encode( "$user:$pass", '' );
    $hdr->header( 'Authorization' => "Basic $authval" );
   }

  my @headers = $treq->get_children_by_tag( 'header' );
  foreach my $h (@headers)
   {
    $hdr->header( $h->{name} => $h->{value} ) if $h->{name};
   }

  if($method eq 'POST')
   {
    my ($body, $type) = extract_body_from_request( $treq );
    $type ||= 'application/x-www-form-urlencoded';
    $hdr->header( 'Content-Type' => $type );

    $req  = new HTTP::Request( $method, $url, $hdr, $body );
   }
  else
   {
    $req  = new HTTP::Request( $method, $url, $hdr );
   }

  my $StartTime = time;
  my $resp = $ua->request( $req );
  $ResponseTime = time - $StartTime;

  my $bValid = validate_response( $test, $resp );

  $bValid ? "ok" : "not ok";
 }



#--------------------------------------------------
#  Test the skip elements to determine if this test/set
#  should be skipped
#
#  Input:  List of skip elements (possibly empty)
#
#  Output: true if we need to skip and false otherwise
sub    handle_skips
 {
  return undef unless @_;
  
  foreach my $skip (@_)
   {
    my $code  = resolve_params( $skip->{if} );
    my $label = exists $skip->{label} ? "'$skip->{label}'" : "($skip->{if})";
    my $ret;
    
    {
     local $SIG{__WARN__} = sub
      {
       my $m = shift;
       $m =~ s/at \(eval \d+\)/in skip $label at/g;
       warn $m;
      };

     $code =~ s/&lt;/</gsm;
     $code =~ s/&gt;/>/gsm;
     $code =~ s/&apos;/'/gsm;
     $ret = eval $code;
    }

   if($@)
    {
     die STOP_EXCEPTION if $@ eq STOP_EXCEPTION;
     $@ =~ s/\(eval \d+\) //g;
     die "skip $label failed: $@\n";
    }

    if($ret)
     {
      my $child = $skip->get_child();
      my $msg = ($child && $child->get_text()) || "skip $label is true.\n";
      printlogvv( "#: $msg" );
      return 1;
     }
   }
  
  return undef;
 }




#--------------------------------------------------
#  Extract the text from the <body/> elements of the
#  supplied request and return a string which is used
#  as the content of the request.
#
#  Input:  $treq  - the request element
#
#  Output: string representing the content of the request
sub    extract_body_from_request
 {
  my $treq = shift;
  # ??? need to modify to deal with multiple body elements
  my @belem = $treq->get_children_by_tag( 'body' );

  return ('', '') unless @belem;
  
  my $body = '';
  my $type = '';

  # ??? change into loop and deal with boundaries
  $body = resolve_params( $belem[0]->get_text_content() );
  $type = $belem[0]->{type} if $belem[0]->{type};

  # ??? need to test for the standard encoding
  if(exists $belem[0]->{encode} and 'yes' eq $belem[0]->{encode})
   {
    $body = uri_escape( $body );
    $body = s/ /\+/g;
   }
  
  ($body, $type);
 }


#sub    extract_body_from_request
# {
#  my $treq = shift;
#  # ??? need to modify to deal with multiple body elements
#  my @belem = $treq->get_children_by_tag( 'body' );
#
#  return ('', '') unless @belem;
#
#  my @bodies = ();
#  my $type   = 1 < @belem ? 'multi-part/mixed' :
#                            ($belem[0]->{type} ? $belem[0]->{type} : '');
#
#  foreach my $bodypart (@belem)
#   {
#    my ($body, $type) = extract_body( $bodypart );
#
#    # Need to handle headers here if necessary.
#    push @bodies, $body;
#   }
#
#  
#  (join_bodies( BOUNDARY, @bodies ), $type);
# }
#
#
#sub  extract_body
# {
#  my $bodypart = shift;
#  
#  my $body = '';
#  my $type = '';
#
#  # ??? change into loop and deal with boundaries
#  $body = resolve_params( $bodypart->get_text_content() );
#  $type = $bodypart->{type} if $bodypart->{type};
#
#  # ??? need to test for the standard encoding
#  if(exists $bodypart->{encode} and 'yes' eq $bodypart->{encode})
#   {
#    $body = uri_escape( $body );
#    $body = s/ /\+/g;
#   }
#
#  ($body, $type);
# }
#
#
#sub  join_bodies
# {
#  my $boundary = shift;
#  my $output   = '';
#  
#  return ($_[0], '' ) if 1 == @_;
#  
#  while(grep { index( $_, $boundary ) != -1 } @_)
#   {
#    $boundary++;
#   }
#
#  return ($output, $boundary);
# }
#


#--------------------------------------------------
#  Validate the HTTP response
#
#  Input:  $test   - reference to the test to execute
#          $resp   - reference to an HTTP::Response object resulting
#                    from an HTTP request
#
#  Output: boolean, true if test is successful, false otherwise
sub  validate_response
 {
  my $test  = shift;
  my $resp  = shift;
  my $tresp = $test->get_child_by_tag( 'response', 0 );

  my $result = validate_and( $tresp, $resp, 'response' );

  unless($result->{success})
   {
    printlogvv( "failed, $result->{'short'}.\n" );
    printlogvv( $result->{'long'}, "\n" )   if $result->{'long'};

    return 0;
   }

  $result->{'success'};
 }


#--------------------------------------------------
sub  is_substr_match
 {
  my $elem = shift;

  defined $elem and (exists $elem->{match} and 'substr' eq $elem->{match});
 }


#--------------------------------------------------
sub  is_regex_match
 {
  my $elem = shift;

  defined $elem and (exists $elem->{match} and 'regexp' eq $elem->{match});
 }


#--------------------------------------------------
sub  is_exist_match
 {
  my $elem = shift;

  defined $elem and (!exists $elem->{match} or 'exist' eq $elem->{match});
 }


#--------------------------------------------------
sub  is_nonexist_match
 {
  my $elem = shift;

  defined $elem and (exists $elem->{match} and 'nonexist' eq $elem->{match});
 }


#--------------------------------------------------
#  Validate a single content test
#
#  Input:  $ct     - reference to a content subtest object
#          $resp   - reference to an HTTP::Response object resulting
#                    from an HTTP request
#
#  Output: reference to a hash containing the result information
sub   validate_content
 {
  my $ct      = shift;
  my $resp    = shift;
  my $content = '';
  my %result  = ( success => 1, short => 'content matches', long => '' );

  $content = resolve_params( $ct->get_text_content() ) || '';

  unless(string_matches( $ct, $resp->content, $content ))
   {
    $result{'short'}   = "content does not match";
    $result{'long'}    = "Expected:\n$content\nReturned:\n" . $resp->content;
    $result{'success'} = 0;
   }

  \%result;
 }


#--------------------------------------------------
#  Perform a string test for a subtest object
#
#  Input:  $ct       - reference to a subtest object
#          $returned - value of a response header
#          $expected - string content expected
#
#  Output: true if match, false otherwise
sub   string_matches
 {
  my $ct       = shift;
  my $returned = shift;
  my $expected = shift;

  is_regex_match($ct)  && return $returned =~ m/$expected/sm;
  is_substr_match($ct) && return -1 < index( $returned, $expected );
  $returned eq $expected;
 }


#--------------------------------------------------
#  Validate a single HTTP header test
#
#  Input:  $hdr    - reference to a header subtest object
#          $resp   - reference to an HTTP::Response object resulting
#                    from an HTTP request
#
#  Output: reference to a hash containing the result information
sub   validate_header
 {
  my $hdr    = shift;
  my $resp   = shift;
  my %result = ( success => 1, short => 'header matches', long => '' );

  my $name   = $hdr->{name};
  my $value  = $hdr->{value};
  my $rethdr = $resp->header( $name );

  if(!defined $rethdr)
   {
    $result{short} = "header \'$name\' is not present";
    $result{success} = 0 unless is_nonexist_match( $hdr );

    return \%result;
   }
  elsif(is_nonexist_match( $hdr ))
   {
    $result{short} = "header \'$name\' is present";
    $result{success} = 0;

    return \%result;
   }

  return \%result if is_exist_match( $hdr );

  unless(string_matches( $hdr, $rethdr, $value ))
   {
    $result{'success'} = 0;
    $result{'short'}   = "header mismatch";
    $result{'long'}    = "expected \'$value\', received \'$rethdr\'";
   }

  \%result;
 }



#--------------------------------------------------
#  Validate a HTTP status code test
#
#  Input:  $st     - reference to a status code subtest object
#          $resp   - reference to an HTTP::Response object resulting
#                    from an HTTP request
#
#  Output: reference to a hash containing the result information
sub   validate_status
 {
  my $st     = shift;
  my $resp   = shift;
  my %result = ( success => 1, short => 'status code matches', long => '' );

  my $code = $st->{code};

  unless(string_matches( $st, $resp->code, $code))
   {
    $result{'success'} = 0;
    $result{'short'}   = 'status code mismatch';
    $result{'long'}    = "expected \'$code\', received \'" . $resp->code . "\'";
   }

  \%result;
 }



#--------------------------------------------------
#  handle group of tests that must all succeed for success
#
#  Input:  $cond   - reference to an and subtest object
#          $resp   - reference to an HTTP::Response object resulting
#                    from an HTTP request
#          $where  - optional string telling where called from
#
#  Output: reference to a hash containing the result information
sub   validate_and
 {
  my $cond   = shift;
  my $resp   = shift;
  my $where  = shift;
  my $result = { success => 1, short => 'tests successful', long => '' };
  my $bShort = defined $where && $where eq 'response';

  foreach my $tst ($cond->get_children())
   {
    next unless $tst->is_element();
    my $res = perform_test( $tst, $resp );

    $result->{'long'} .= "\n -- and --\n"  if $result->{'long'};
    $result->{'long'} .= $res->{'short'};
    $result->{'long'} .= "\n".$res->{'long'} if $res->{'long'};
    $result->{'success'} &&= $res->{'success'};

    return $bShort ? $res : $result if !$result->{'success'};
   }

  $result;
 }



#--------------------------------------------------
#  handle group of tests that must all succeed for success
#  all tests are always performed 
#
#  Input:  $cond   - reference to an and subtest object
#          $resp   - reference to an HTTP::Response object resulting
#                    from an HTTP request
#
#  Output: reference to a hash containing the result information
sub   validate_and_all
 {
  my $cond   = shift;
  my $resp   = shift;
  my $result = { success => 1, short => 'tests successful', long => '' };

  foreach my $tst ($cond->get_children())
   {
    next unless $tst->is_element();
    my $res = perform_test( $tst, $resp );

    $result->{'long'} .= "\n -- and --\n"  if $result->{'long'};
    $result->{'long'} .= $res->{'short'};
    $result->{'long'} .= "\n".$res->{'long'} if $res->{'long'};
    $result->{'success'} &&= $res->{'success'};
   }

  $result;
 }



#--------------------------------------------------
#  handle group of tests that one must succeed for success
#
#  Input:  $cond   - reference to an or subtest object
#          $resp   - reference to an HTTP::Response object resulting
#                    from an HTTP request
#
#  Output: reference to a hash containing the result information
sub   validate_or
 {
  my $cond   = shift;
  my $resp   = shift;
  my $result = { success => 0, short => 'tests failed', long => '' };

  foreach my $tst ($cond->get_children())
   {
    next unless $tst->is_element();
    my $res = perform_test( $tst, $resp );

    $result->{'long'} .= "\n -- or --\n" if $result->{'long'};
    $result->{'long'} .= $res->{'short'};
    $result->{'long'} .= "\n".$res->{'long'} if $res->{'long'};
    $result->{'success'} ||= $res->{'success'};

    return $result if $result->{'success'};
   }

  $result;
 }



#--------------------------------------------------
#  handle group of tests that one must succeed for success
#  all tests are always performed 
#
#  Input:  $cond   - reference to an or subtest object
#          $resp   - reference to an HTTP::Response object resulting
#                    from an HTTP request
#
#  Output: reference to a hash containing the result information
sub   validate_or_all
 {
  my $cond   = shift;
  my $resp   = shift;
  my $result = { success => 0, short => 'tests failed', long => '' };

  foreach my $tst ($cond->get_children())
   {
    next unless $tst->is_element();
    my $res = perform_test( $tst, $resp );

    $result->{'long'} .= "\n -- or --\n" if $result->{'long'};
    $result->{'long'} .= $res->{'short'};
    $result->{'long'} .= "\n".$res->{'long'} if $res->{'long'};
    $result->{'success'} ||= $res->{'success'};
   }

  $result;
 }



#--------------------------------------------------
#  handle a test that must fail for success
#
#  Input:  $not    - reference to a not subtest object
#          $resp   - reference to an HTTP::Response object resulting
#                    from an HTTP request
#
#  Output: reference to a hash containing the result information
sub   validate_not
 {
  my $not    = shift;
  my $resp   = shift;
  my $result = undef;

  my $tst = (grep { $_->is_element() } $not->get_children())[0];

  $result = perform_test( $tst, $resp );
  $result->{'long'}    = "not ($result->{'long'})" if $result->{'long'};
  $result->{'short'}   = "not ($result->{'short'})";
  $result->{'success'} = !$result->{'success'};

  $result;
 }


#--------------------------------------------------
#  test response time of the request
#
#  Input:  $tst    - reference to a time subtest object
#          $resp   - reference to an HTTP::Response object resulting
#                    from an HTTP request
#
#  Output: reference to a hash containing the result information
sub  validate_time
 {
  my $tst      = shift;
  my $resp     = shift;
  my %result   = ( success => 1, short => 'response time in range', long => '' );
  my $resptime = response_time();  

  if(exists $tst->{min} and $tst->{min} and $resptime < $tst->{min})
   {
    $result{'short'}   = "response time shorter than expected";
    $result{'long'}    = "Expected: $tst->{min} secs, returned: $resptime secs";
    $result{'success'} = 0;
   }

  if(exists $tst->{max} and $tst->{max} and $resptime > $tst->{max})
   {
    $result{'short'}   = "response time longer than expected";
    $result{'long'}    = "Expected: $tst->{max} secs, returned: $resptime secs";
    $result{'success'} = 0;
   }

  \%result;
 }



#--------------------------------------------------
#  execute code to test response
#
#  Input:  $tst    - reference to a code subtest object
#          $resp   - reference to an HTTP::Response object resulting
#                    from an HTTP request
#
#  Output: reference to a hash containing the result information
sub  validate_code
 {
  my $tst     = shift;
  my $resp    = shift;
  my $label   = $tst->{label} || '';
  my %result  = ( success => 1, short => "code test '$label' successful", long => '' );
  my @ret     = (1, '', '');

  my $code = resolve_params(  $tst->get_text_content() ) || '';

  {
   local $SIG{__WARN__} = sub
    {
     my $m = shift;
     $m =~ s/at \(eval \d+\)/in code test '$label' at/g;
     warn $m;
    };

   $code =~ s/&lt;/</gsm;
   $code =~ s/&gt;/>/gsm;
   $code =~ s/&apos;/'/gsm;
   @ret = eval $code;
  }

  if($@)
   {
    die STOP_EXCEPTION if $@ eq STOP_EXCEPTION;
    $@ =~ s/\(eval \d+\) //g;
    die "Code($label) failed: $@\n";
   }

  $result{success} = $ret[0];
  $result{short}   = $ret[1] || ($ret[0]
                                  ? "code test '$label' successful"
                                  : "code test '$label' failed"
                                 );
  $result{long}    = $ret[2] || '';

  \%result;
 }



#--------------------------------------------------
#  pseudo-subtest for sending messages or aborting
#  on a really critical error.
#
#  Input:  $tst    - reference to a msg pseudo-subtest object
#          $resp   - reference to an HTTP::Response object resulting
#                    from an HTTP request
#
#  Output: reference to a hash containing the result information
sub   send_msg
 {
  my $tst     = shift;
  my $resp    = shift;
  my %result  = ( success => exists $tst->{fail} ? 'no' eq $tst->{fail} : 0,
                  short => 'msg encountered',
                  long  => '' );

  $result{success} = msg( $tst->get_text_content(),
                          $tst->{print}, $tst->{action}
                        );
                        
  \%result;
 }




#--------------------------------------------------
#  Supply the actual function of the msg pseudo-subtest.
#  Print the supplied message under the appropriate condition, and
#  take the supplied action.
#
#  Input:  $msg    - message to display
#          $print  - specify conditions under which to print
#          $action - specify action to take after printing
#
#  Output: a boolean telling if the action was success or fail
sub msg
 {
  my $msg    = shift || '';
  my $print  = shift || 'normal';
  my $action = shift || 'fail';
  
  $msg = resolve_params( $msg );
  $msg =~ s/[ \t]+$//s;

  PRINT:
   {
    'always'  eq $print && do { printlog( $msg );   last PRINT; };
    'normal'  eq $print && do { printlogv( $msg );  last PRINT; };
    'verbose' eq $print && do { printlogvv( $msg ); last PRINT; };
   };

  ACTION:
   {
    'success' eq $action && do { return 1;           };
    'fail'    eq $action && do { return 0;           };
    'stop'    eq $action && do { die STOP_EXCEPTION; };
    'exit'    eq $action && do { exit( 1 );          };
   };
 }




#--------------------------------------------------
#  pseudo-subtest for filtering the HTTP response.
#
#  Input:  $tst    - reference to a filter pseudo-subtest object
#          $resp   - reference to an HTTP::Response object resulting
#                    from an HTTP request
#
#  Output: reference to a hash containing the result information
sub  execute_filter
 {
  my $tst    = shift;
  my $resp   = shift;
  my $label  = $tst->{label} || '';
  
  my $code = resolve_params( $tst->get_text_content() ) || '';

  {
   local $SIG{__WARN__} = sub
    {
     my $m = shift;
     $m =~ s/at \(eval \d+\)/in filter '$label' at/g;
     warn $m;
    };

   $code =~ s/&lt;/</gsm;
   $code =~ s/&gt;/>/gsm;
   $code =~ s/&apos;/'/gsm;
   if(exists $tst->{class} and 'response' eq $tst->{class})
    {
     eval $code;
    }
   else
    {
     local $_ = $resp->content;
     eval $code;
     $resp->content( $_ );
    }
  }

  if($@)
   {
    die STOP_EXCEPTION if $@ eq STOP_EXCEPTION;
    $@ =~ s/\(eval \d+\) //g;
    die "filter($label) failed: $@\n";
   }
   
  { success => 1, short => 'executing filter', long  => '' };
 }



sub   call_subtest_definition
 {
  my $tst    = shift;
  my $resp   = shift;
  
  my $ref = $tst->{subtest};
  die "call element missing subtest attribute\n" unless defined $ref;
  die "No subtest-def referenced by '$ref'.\n" unless exists $SubTests{$ref};

  validate_and( $SubTests{$ref}, $resp, 'call' );
 }


#--------------------------------------------------
#  Log the appropriate message for the current subtest
#
#  Input:  $tag    - the subtest type
#          $test   - reference to a subtest object
#          
sub  log_test
 {
  my $tag  = shift;
  my $test = shift;
  my $label;
  
  if($tag eq 'filter')
   {
    $label = "Filter: " . ($test->{label} || '') . "\n";
   }
  elsif($tag eq 'msg')
   {
    $label = "Msg:\n";
   }
  elsif($tag eq 'call')
   {
    my $id   = $test->{subtest};
    my $def  = $SubTests{$id};
    my $name = defined $def && $def->{label} ? $def->{label} : "($id)";
    $label = "Call: $name\n";
   }
  else
   {
    $label = "Test: " . ($test->{label} || $tag) . "\n";
   }

  printlogvv( $label );
 }


#--------------------------------------------------
#  Vector to the correct kind of test
#
#  Input:  $test   - reference to a subtest object
#          $resp   - reference to an HTTP::Response object resulting
#                    from an HTTP request
#
#  Output: reference to a hash containing the result information
sub  perform_test
 {
  my $test = shift;
  my $resp = shift;
  my $tag  = $test->tagname();

  log_test( $tag, $test );

  $Tests{$tag}->( $test, $resp );
 }


#--------------------------------------------------
#  Read parameters from the battery and
#   command line
#
#   Input:  $ts  - reference to a battery object
sub  setup_params
 {
  my $ts = shift;
  my $param;

  foreach $param ($ts->get_children_by_tag( 'param' ))
   {
    my $name  = $param->{name};
    die "Invalid param name \'$name\'.\n" unless $name =~ /^\w+$/;

    my $value = $param->{value};

    $value = $CmdLine{$name} if exists $CmdLine{$name};

    param( $name => resolve_params( $value ) );
   }

  foreach $param (keys %CmdLine)
   {
    param( $param => $CmdLine{$param} );
   }
 }


#--------------------------------------------------
#  Replace all parameters found in the supplied
#   string with the value of those parameters.
#
#   Input:   $str  - string possibly containing replaceable parameters
#
#   Output:  string with replaceable parameters resolved
sub  resolve_params
 {
  my $str = shift;

  if($str)
   {   # If param replace with param value, otherwise exec Perl expr.
    $str =~ s/{{([^}]+)}}/expand_a_param( $1 )/smeg;
   }

  $str;
 }


#-------------------------------------------------
#  Take a potential parameter reference and expand
#  it either as the value of the parameter or as a
#  Perl expression.
#  
#  Input:  $parmref  - a suspected parameter reference
#  
#  Output: the expanded text of the parameter or
#          result of the Perl expression
sub   expand_a_param
 {
  my $parmref = shift;

  unless($parmref =~ /^\w+$/)      # doesn't appear to be a parameter
   {
    return eval $parmref;
   }
  elsif(param_exists( $parmref ))
   {
    return param( $parmref );
   }
  else                             # looks like a parameter, it's not
   {
    die "Reference of unknown parameter \'$parmref\'\n";
   }
 }



#--------------------------------------------------
#  get/set parameter values
sub   param
 {
  my $name = shift;
  
  $Params{$name} = shift if @_;
  
  $Params{$name};
 }


#--------------------------------------------------
#  Check for existence of a parameter
sub   param_exists
 {
  my $name = shift;

  exists $Params{$name};
 }


#--------------------------------------------------
#  Initialize the Parameters storage
sub   params_initialize
 {
  %Params = ();
 }


#--------------------------------------------------
#  get/set option values
sub   option
 {
  my $name = shift;
  
  $Options{$name} = shift if @_;
  
  $Options{$name};
 }


#--------------------------------------------------
#  Check for existence of an option
sub   option_exists
 {
  my $name = shift;

  exists $Options{$name};
 }



#--------------------------------------------------
#  Return the time needed for the request to complete.
sub   response_time
 {
  $ResponseTime;
 }



#--------------------------------------------------
#  Extract parameters of the form 'name=value' from
#   the command line.
sub   extract_cmd_line
 {
  my @args=();

  foreach(@ARGV)
   {
    if(/^([^=]+)=([^=]*)$/)
     {
      $CmdLine{$1} = $2;
     }
    else
     {
      push @args, $_;
     }
   }

  @ARGV = @args;
 }



#--------------------------------------------------
#  If the -f command line option has been used to
#  supply a test suite file listing batteries, expand
#  that file onto the command line.
sub    expand_list_file
 {
  return unless defined $opts{f} and -e $opts{f};

  local *L;
  local $_;
  open( L, $opts{f} ) or die "Unable to open suite file '$opts{f}'.\n";
  my @batteries = ();
  while(<L>)
   {
    s/#.*//;
    next if /^\s*$/;

    push @batteries, split( /\s+/, $_ );
   }
  close( L );

  unshift @ARGV, @batteries;
 }




#---------------------------------------------------
#  Compile library code which may be used elsewhere
#  in a test.
#
#  Input:  $code    - reference to a codelib to compile
#          $Counts  - reference to the counts structure
sub   execute_a_test_set
 {
  my $set    = shift;
  my $Counts = shift;
  my $label  = $set->{label} || 'unlabeled test-set';

  # Startup a test-set
  printlogv( "# Starting '$label'\n" );

  if(handle_skips( $set->get_children_by_tag( 'skip' ) ))
   {
    my $skips = count_tests( $set );
    $Counts->{skip} += $skips;
    printlogv( "# Skipping $skips tests\n" );
   }
  else
   {
    # ignore skip elements, we've already processed them.
    my @children = grep { 'skip' ne $_->tagname() } $set->get_children();
    process_set_of_tests( $Counts, @children );
   }

  # Cleanup the test-set
  printlogv( "# Ending '$label'\n" );
 }


#---------------------------------------------------
#  Compile library code which may be used elsewhere
#  in a test.
#
#  Input:  $code  - reference to a codelib to compile
sub   compile_lib
 {
  my $code  = shift;
  my $label = $code->{label};
  $label ||= defined $code->{id} ? "($code->{id})" : 'unknown';

  printlogvv( "Compile codelib '$label'\n" );

  if(exists $code->{id})
   {
    my $id = $code->{id};

    if(exists $CodeLibIDs{$id})
     {
      printlogvv( "codelib ($id) already compiled\n" );
      return 1;
     }

    $CodeLibIDs{$id} = 1;
   }

  {
   # signal handler to clean up compile warnings
   local $SIG{__WARN__} = sub
    {
     my $m = shift;
     $m =~ s/at \(eval \d+\)/in codelib '$label' at/g;
     warn $m;
    };

   my $content = resolve_params( $code->get_text_content() );
   $content =~ s/&lt;/</gsm;
   $content =~ s/&gt;/>/gsm;
   $content =~ s/&apos;/'/gsm;
   eval $content;
  }

  if($@)
   {
    die STOP_EXCEPTION if $@ eq STOP_EXCEPTION;
    $@ =~ s/\(eval \d+\) //g;
    die "Codelib($label) failed: $@\n" ;
   }
 }



#--------------------------------------------------------
#  Create a callable subtest definition
#
#  Input:  $subtest  - reference to a subtest definition
sub  build_subtest_definition
 {
  my $subtest = shift;
  my $label = ($subtest->{label} || "($subtest->{id})");
  
  my $id = $subtest->{id};
  die "Missing id attribute for subtest-def '$label'.\n" unless defined $id;
  die "Subtest definition '$id' already defined.\n" if exists $SubTests{$id};
  
  $SubTests{$id} = $subtest;
 }



#--------------------------------------------------------
#  Search the current directory and include path to find
#  an include file
#
#  Input:  $file   - name of file to find
#
#  Output: filename qualified with the appropriate path.
#          If the filename is not found, the original name
#          is returned
sub   find_include_file
 {
  my $file = shift;

  return $file if File::Spec->file_name_is_absolute( $file ) or -e $file;

  foreach my $dir (@{option( 'incpath' )})
   {
    my $f = File::Spec->catfile( $dir, $file );

    return $f if -e $f;
   }

  return $file;
 }



#---------------------------------------------------------------
#  Return the content of the root element in the file referenced
#  by the supplied include element. Recursively expand includes
#  in the children.
#
#  Input:   $inc  - reference to an include object
#
#  Output:  one or more xml objects to replace the include object
sub   expand_an_include
 {
  my $inc  = shift;
  my $file = find_include_file( $inc->{file} );

  printlogvv( "Including ", ($inc->{label} || $file || 'unkown'), "\n" );
  die "Missing or empty file attribute.\n" unless $file;

  if(exists $inc->{parse} and 'text' eq $inc->{parse})
   {
    local *I;
    local $/ = undef;
    open( I, $file ) or die "Unable to open include file '$file'.\n";
    my $content = <I>;
    close( I );
    XMLdata::make_text( $content );
   }
  else
   {
    my $content = $parser->parsefile( $file );
    die "Invalid include file.\n" unless defined $content;

    my $root = XMLdata::root( $content );
    die "Invalid include file.\n" unless defined $root;
    warn "Include file root not an inclusion element.\n"
                                 unless 'inclusion' eq  $root->tagname();

    expand_includes( $root );

    $root->get_children();
   }
 }


#---------------------------------------------------------------
#  Walk the supplied container element, replacing all include
#  elements with their contents.
#
#  Input:  $contain   - a container object in which we will expand
#                       include elements
sub   expand_includes
 {
  my $contain = shift;

  $contain->{Kids} = [ map { 'include' eq  $_->tagname() ?
                                  expand_an_include( $_ ) : ($_ )
                           }
                       @{$contain->{Kids}}
                     ];
 }



#----------------------------------------------------------------
#  Check the test set for validity.
#
#  Input:  $battery  - reference to a battery object
#
#  Output: false if valid, a string describing the problem if not
sub   invalid_battery
 {
  my $battery = shift;
  my %LegalTags = map { $_ => 1 } @TopLevelElements, @TestComponents;

  my $tag = $battery->tagname();

  return "Invalid root element." unless 'battery' eq $tag or 'testset' eq $tag;

  my $is_valid = sub { my $el = shift;
                       die $el->tagname()."\n" 
                                    unless !$el->is_element() or
                                    $LegalTags{$el->tagname()};
                     };

  eval
   {
    XMLdata::visit_elements( $battery, $is_valid );
   };

  if($@)
   {
    chomp( $@ );
    return "Unrecognized element type \'$@\'";
   }
  else
   {
    return 0;
   }
 }


#---------------------------------------------------
#  Open the logging file, if necessary
sub   open_logfile
 {
  if(option( 'logging' ))
   {
    my $logfile = option( 'logfile' );
    open( LOG, ">>$logfile" ) or die "Unable to open log file '$logfile'.\n";
   }
 }



#---------------------------------------------------
#  Close the logging file, if necessary
sub  close_logfile
 {
  close( LOG ) if option( 'logging' );
 }



#---------------------------------------------------
#  Print normal messages and add to log.
#
#  Input:  A list of messages to print
sub    printlog
 {
  print @_;
  print LOG @_ if option( 'logging' );
 }



#---------------------------------------------------
#  Print normal verbose messages and add to log.
#
#  Input:  A list of messages to print
sub    printlogv
 {
  print @_     if 'quiet' ne option( 'msgs' );
  print LOG @_ if option( 'logging' );
 }



#---------------------------------------------------
#  Print extra verbose messages and add to log.
#
#  Input:  A list of messages to print
sub  printlogvv
 {
  print @_     if 'verbose' eq option( 'msgs' );
  print LOG @_ if option( 'logging' );
 }



#--------------------------------------------------
#  Usage message
sub  usage
 {
  die <<"EOU";
Usage: $0 [-vwqVl] [-o logfile] [-I incpath] [-f suite] [-jJ cjar] [name=value ...] test.files ...

Where:  -v   verbose output
        -w   normal output, default
        -q   quiet output
        -V   print version information
        -l   log results to HTTPtest.log
        -o   send results to specified logfile
        -I   semicolon-separated path to check for include files
        -f   supply a test suite filename
        -j   use a temporary cookie jar
        -J   use the cookie jar file 'cjar'

name=value pairs override parameters in the test files

test.files is a list of files that contain the tests to run
EOU
 }

__END__

=head1 NAME

HTTPtest.pl

=head1 Purpose

B<HTTPtest> is a generalized HTTP request test program. It allows for
efficient regressions testing of any HTTP-based server.

B<Warning:> This program is under construction, so the documentation is
subject to change.

=head1 SCRIPT CATEGORIES

HTTP - suggested

=head1 PREREQUISITES

This script depends on both the C<strict> and C<vars> pragmas. The script also uses
the C<FindBin> C<LWP::UserAgent>, C<HTTP::Cookies>, C<URI::Escape>, C<XML::Parser>,
C<Getopt::Std>, and C<File::Spec> modules.

=head1 COREQUISITES

If the C<Time::HiRes> module is available, it is used to generate higher-resolution
timings on the time subtest and the script timing.

=head1 OSNAMES

any. Tested on MSWin32 and Linux.

=head1 Description

The B<HTTPtest> program uses a battery file to describe a series of tests
to perform using the HTTP protocol. In general, a set of tests are aimed
at a single server or service. Like HTTP itself, the tests are defined
in terms of a request and a response. The battery file describes the tests
in an XML-based format, described below.

B<HTTPtest> reads each test in the battery file, performs the specified request,
and evaluates the response based on the criteria described in the test. If
all criteria are met, the test is successful. Otherwise, the test fails.

=head1 Usage

B<HTTPtest> is executed as follows:

 perl HTTPtest.pl [-vwqVl] [-o logfile] [-I incpath] [-f suite] [-jJ cjar] [name=value ...] test.files ...

   Where:  -v   verbose output
           -w   normal output, default
           -q   quiet output
           -V   print version information
           -l   log results to HTTPtest.log
           -o   send results to specified logfile
           -I   semicolon-separated path to check for include files
           -f   supply a test suite filename
           -j   use a temporary cookie jar
           -J   use the cookie jar file 'cjar'

   name=value pairs override parameters in the test files

   test.files is a list of files that contain the tests to run

If run with a bad option or no test files, the above message is printed.

=head1 Battery File

The battery file contains a C<battery> root element. The C<battery> may
contain C<include> elements to include information from other files,
C<param> elements which factor out reused parameters, and C<test> elements
which describe the tests to run.

Each C<test> consist of a C<request>, which describes the HTTP request, and
a C<response> which describes the criteria for evaluating the HTTP response.
A C<request> specifies the URL to call including any query parameters, an
HTTP method such as I<GET> or I<POST>, and an optional request body.

A C<response> contains a series of subtests which include the ability to test:

=over 4

=item status

The response status code for the response. Reasonable values include I<200>
and I<404>.

=item header

Tests for the existence or value of particular MIME-headers in the response.
Useful for verifying redirections and testing returned cookies.

=item content

Evaluation subtest for the content of response.

=item time

Compare the response time of the URL request against a minimum and maximum
timespan in seconds.

=item code

Execute arbitrary Perl code on the response to evaluate the response.

=item msg

This pseudo-subtest displays a message to the user and optionally aborts
the current battery or program run.

=item filter

This pseudo-subtest executes Perl code specifically designed to perform
some form of filtering on the response. Always succeeds.

=item call

This pseudo-subtest executes a list of subtests defined using the
C<subtest-def> element.

=back

By default all of the evaluations in the C<response> must succeed in order
for the C<test> to succeed. However, five other elements are supported to
allow for arbitrary logical combinations of the evaluations.

=over 4

=item and

All of the subtests contained in the C<and> element must succeed in order for
the C<and> to succeed. Any unsuccessful evaluation causes the C<and> to fail
without executing any remaining children.

=item and-all

All of the subtests contained in the C<and-all> element must succeed in order
for the C<and-all> to succeed. Unlike the C<and> subtest, all children are
executed regardless of the success or failure of any child.

=item or

One of the subtests contained in the C<or> element must succeed in order for
the C<or> to succeed. Any successful evaluation causes the C<or> to succeed
without executing any remaining children.

=item or-all

One of the subtests contained in the C<or-all> element must succeed in order
for the C<or-all> to succeed. Unlike the C<or> subtest, all children are
executed regardless of the success or failure of any child.

=item not

The sense of the subtest contained in the C<not> element is reversed, so that
a failure becomes a success.

=back

=head1 Element List

=head2 and

The C<and> element is used to combine a set of subtests in a logical I<and>
relationship. The C<and> element can contain any of the subtest elements or
the logical elements. The C<and> element has one optional attribute I<label>
which is a displayable description of the set of subtests contained in the
C<and>. It may be used in logging and failure reporting.

B<HTTPtest> evaluates the response based on each of the children of the
C<and> in turn. When a subtest fails, no more children are tested and the
C<and> evaluation fails.

=head2 and-all

The C<and-all> element is used to combine a set of subtests in a logical I<and>
relationship. The C<and-all> element can contain any of the subtest elements or
the logical elements. The C<and-all> element has one optional attribute I<label>
which is a displayable description of the set of subtests contained in the
C<and-all>. It may be used in logging and failure reporting.

B<HTTPtest> evaluates the response based on each of the children of the
C<and-all> in turn.  Unlike the C<and> subtest, all of the children are
evaluated regardless of the success or failure of any given child.

=head2 auth

The C<auth> element supplies an optional user id and password for HTTP
authentication of a request. Currently only supports I<Basic> authentication.
The C<auth> element has two reuired attributes I<user> and I<password>. These
attributes have the obvious values.

=head2 battery

The C<battery> is the root element of the battery file. It can contain zero
or more C<param> elements, followed by zero or more C<codelib> elements, followed
by zero or more C<test> elements. In addition, C<include> elements may occur
anywhere in the C<battery> element.

The C<battery> has one required attribute, I<label> which is a short displayable
description of the test set.

=head2 body

The C<body> element supplies the content for an HTTP request. It is currently
only supported for the I<POST> HTTP method. The content of this element is
used as the body of the request by B<HTTPtest>. Since this content resides in
an XML document, it may be necessary to wrap the content in a CDATA section
in order for the file to remain well-formed.

The C<body> element has two optional attributes. The C<type> attribute specifies
the MIME type of the content. By default, the content type is
I<application/x-www-form-urlencoded>, which is the type used by browsers on HTML
forms. The C<encode> attribute specifies whether the content needs to be
URL-encoded. The default value is I<no>. If you have already encoded the data, the
correct answer is I<no>.

=head2 call

The C<call> element executes the list of subtests that have been defined in
an earlier C<subtest_def> element. This element allows a group of subtests to
be factored out of a C<response> and reused in other C<response>s.

=head2 code

The C<code> element specifies a piece of Perl code to execute in order to evaluate
the HTTP response. The C<code> element has one optional attribute. The C<label>
attribute provides a displayable description of this element that can be used in
error reporting and logging.

The content of this element is raw Perl code. Since Perl code can easily violate the
well-formedness criteria, it is recommended that the whole content of this element be
placed in a CDATA section.

The Perl code is executed with a set of variables that are accessible. These variables
can be used to determine how to perform the tests or communicate with other pieces
of code. The variables of interest are:

=over 4

=item $resp

This is a reference to an HTTP::Response object containing the current response. This
is a C<my> variable, and therefore only available directly inside the C<code> element.

=item $label

This string is the value of the I<label> attribute for the current test. This is a C<my>
variable, and therefore only available directly inside the C<code> element.

=item %Params

This is a global hash matching all of the parameter names to their current values. The
Perl code is allowed to modify values in this hash to pass data to other tests.
Direct access to this hash is deprecated. Parameters should be accessed through
the following methods:

=over 4

=item param()

This is the get/set function for parameters. When called with one argument,
the value of the parameter with the supplied name is returned. If the function
is called with two arguments, the parameter named by the first argument is set
to the value of the second argument. For example,

    my $server = param( 'server' );
    param( 'server' => 'www.example.com' );

=item param_exists()

This function returns a boolean value that specifies whether the named parameter
exists. It has one argument, the name of the parameter to test.

=back

=item $parser

This is a global reference to an XML::Parser object configured with the I<Objects> style.
This object may be used to parse XML from the response to simplify testing. See the
XMLdata.pm library file for tools that can manipulate this structure.

=item response_time()

This function returns the time in seconds for the HTTPresponse to return.

=back

The Perl code should return a list of scalars. The first item in the list is the
only required item. It is true if the test was successful, and false otherwise. The
second item in the list is a short string describing the result of the test. The third
and final item is a long string giving a verbose description of the result of the test.
If either the second or third items is missing, a generic answer is used in each case.

=head2 codelib

The C<codelib> element declares Perl code that is to be compiled into the test script
for use in later tests. This is a good place to use modules or define routines that
are needed by C<code> subtest defined later in the file. The content of this element
is raw Perl code. Since Perl code can easily violate the well-formedness criteria, it
is recommended that the whole content of this element be placed in a CDATA section.

The C<codelib> element has one required attribute and one optional attribute. The I<label>
attribute provides a displayable description of this element that can be used in error
reporting and logging. The optional I<id> attribute is used to uniquely identify a
C<codelib>. This helps B<HTTPtest> to recognize if a C<codelib> is encountered more than
once. (For example through the use of C<include> elements.) If a C<codelib> I<id> is
encountered a second time, that C<codelib> is not evaluated the second or subsequent times.
If a C<codelib> does not contain an I<id> and it is encountered more than once, it is
evaluated each time it is encountered.

=head2 content

The C<content> element specifies subtest for the body of the HTTP response.
The C<content> element has one optional attribute. If the I<match> attribute is missing
or has a value of I<exact>, the body of the HTTP response must match the content of the
C<content> element exactly. If the I<match> attribute has a value of I<regexp>, the
content of the element is treated as a regular expression to be matched against the
HTTP response content. If the I<match> attribute has a value of I<substr>, the
content of the element is expected to exist somewhere in the response.

B<WARNING:> There is a limit to the size of a regular expression in Perl. It is better to
break a complicated test into several smaller tests.

=head2 filter

The C<filter> element is a special subset of the functionality of the C<code>
element. The C<filter> element has two optional attributes. The I<label>
attribute supplies a displayable label for this filter. The I<class> attribute
specifies which class of filter we are specifying. If the I<class> attribute
is missing or if it has the value I<content>, the filter modifies the content
of the response which is available in the variable C<$_>. If the I<class>
attribute has the value I<response>, the filter can affect any portion of the
response. The response is available in the variable C<$resp> as an
C<HTTP::Response> object.

=head2 header

The C<header> element specifies a MIME header in either a C<request> or a
C<response>. It has three attributes that are treated slightly differently
depending on whether the element is contained in a C<request> or a C<response>.
In both cases, the I<name> attribute is required and supplies the name of
the MIME header of interest.

The I<value> attribute specifies the value of that MIME header. In the case of
a C<request>, this attribute is required and specifies the literal text of the
value of the header. In the case of a C<response>, the I<value> attribute is
interpreted as described by the I<match> attribute.

The I<match> attribute is specifies the interpretation of the I<value> attribute
on a C<header> in a C<response>. The default value of the I<match> attribute
is I<exact>. The valid values for I<match> are

=over 4

=item exact

The value of the MIME header must exactly match the I<value> of the C<header>.

=item regexp

The I<value> of the C<header> is treated as a regular expression to match the value
of the specified MIME header.

=item exist

The MIME header must exist in the HTTP response. The I<value> attribute is ignored.

=item nonexist

The MIME header must not exist in the HTTP response. The I<value> attribute is ignored.

=back

=head2 include

The C<include> element allows portions of a C<battery> to be factored into separate
files. This supports sharing of information between battery files and can simplify
maintenance of test sets.

The C<include> element has one required attribute (I<file>) and two optional attributes
(I<label> and I<parse>). The I<file> attribute specifies the file to be included.
B<HTTPtest> looks for the file as it is specified, relative to the current directory, first.
If the file is not found and the filename is not an absolute path, B<HTTPtest> then attempts
to find the file by appending the I<file> attribute to each path in the include path.
I<label> attribute is a displayable description only used for logging and error reporting.

The I<parse> attribute tells B<HTTPtest> how to treat the data from the included file. If
the I<parse> attribute is missing or if its value is I<xml>, the contents of the included
file are parsed and all children of the root element replace the C<include> element. If the
value of the I<parse> attribute is I<text>, the contents of the file are treated as a text
node that replaces the C<include> element. Since this data is not handled directly by the
XML parser, there is no need to use character entities such as C<&lt;> and C<&amp;>.

One extra constraint on the C<include> is that the test set file should still be a valid
document after the C<include> elements are replaced by the content of their files.

=head2 inclusion

The C<inclusion> element is the root element of any included I<xml>-parsed file. Its purpose
is to provide a well-formed XML document for the included file.

=head2 msg

The C<msg> element is used to report unusual problems to the user and optionally to
abort a test run. The C<msg> element has two optional attributes and contains text
to display. The I<print> attribute determines when the message is printed. Possible
values are:

=over 4

=item always

Always print the message when the C<msg> element is encountered.

=item normal

Print the message when the C<msg> element is encountered and the B<-w> or B<-v> switches
are active. This is the default printing behavior.

=item verbose

Print the message when the C<msg> element is encountered and the B<-v> switch
is active.

=back

The I<action> attribute determines the effect of this pseudo-subtest on the test
in progress. The possible values are:

=over 4

=item success

Treat the C<msg> element as a successful subtest.

=item fail

Treat the C<msg> element as a failed subtest. This is the default action.

=item stop

Stop the current battery, but continue processing any battery files remaining on
the command line.

=item exit

Exit the HTTPtest.pl script, altogether.

=back

The C<msg> element is best used as part of an C<or> or C<and> condition. This
allows an earlier subtest to determine whether or not the C<msg> element is
processed at all.

=head2 not

The C<not> element is used to logically invert the outcome of a subtest.
The C<not> element can contain any one of the subtest elements or the logical elements.
The C<not> element has one optional attribute I<label> which is a displayable description.
It may be used in logging and failure reporting.

=head2 option

The C<option> element configures default processing options for B<HTTPtest>. All of the
options configurable through the C<option> element are also available through command
line parameters. The command line parameters override whatever C<option>s are in the test
set file. All of the C<option> elements must come before any other elements in
the C<battery>.

The C<option> element has two required attributes, I<name> and I<value>. The I<value>
attributes will be interpreted differently depending on option they are applied to. The
allowable options are:

=over 4

=item incpath

The I<value> attribute for this option is a semicolon-separated list of paths to use in finding
include files. Unlike the other options, the command line switch does not override this option.
Instead the paths supplied by the B<-I> switch are prepended to the include path.

=item logging

The I<value> for this option is either I<yes> (enable logging) or I<no> (disable logging). This
is equivalent to the B<-l> switch.

=item logfile

The I<value> for this option is the name of a file into which the logging information is written.
This is equivalent to the B<-o> switch.

=item msgs

The I<value> for this option is either an empty string (normal output), I<quiet> (quiet mode),
or I<verbose> (verbose mode). This modes are equivalent to the I<-w>, I<-q>, and I<-v> switches.

=item cookiejar

The I<value> for this option is either an empty string (temporary cookie jar) or the name of a
file to store the cookie jar. If this option is supplied, HTTPtest retains cookies across all
of the C<test>s in the C<battery>. If no name is specified, the cookie jar will be empty when
the script begins and all cookies wil be discarded at exit. This usage is equivalent to the
B<-j> switch. If a filename is specified, that file is read to initialize the cookie jar if the
file exists. When the program prepares to exit, the current state of the cookie jar is saved
to the specified file. This usage is equivalent to the B<-J> switch.

=back

=head2 or

The C<or> element is used to combine a set of subtests in a logical I<or>
relationship. The C<or> element can contain any of the subtest elements or the
logical elements. The C<or> element has one optional attribute I<label> which
is a displayable description of the set of subtests contained in the C<or>. It
may be used in logging and failure reporting.

B<HTTPtest> evaluates the response based on each of the children of the
C<or> in turn. When an subtest succeeds, no more children are tested and the
C<or> subtest succeeds.

=head2 or-all

The C<or-all> element is used to combine a set of subtests in a logical I<or>
relationship. The C<or-all> element can contain any of the subtest elements or
the logical elements. The C<or-all> element has one optional attribute I<label>
which is a displayable description of the set of subtests contained in the
C<or-all>. It may be used in logging and failure reporting.

B<HTTPtest> evaluates the response based on each of the children of the
C<or-all> in turn. Unlike the C<or> subtest, all of the children are evaluated
regardless of the success or failure of any given child.

=head2 param

The C<param> element is an empty element which gives a name to a piece of text
which is expected to be reused many times in the test set. A good example would
be the name of the server undergoing test. For example,

  E<lt>param name="server" value="testserver.telescan.com"/>

Using a C<param> in this way simplifies pointing the test to a new server. The
C<param> is then referenced in any text or attribute in the test set file using
the form I<{{server}}>.

In addition, the value of a C<param> may be changed from the B<HTTPtest> command
line using the syntax: I<name=value>. To run this C<battery> on the server
I<live.telescan.com>, you would add I<server=live.telescan.com> to the command
line before the name of the battery file.

The C<param> element has two required attributes. The I<name> attribute specifies
the name by which this parameter is referenced. The I<value> attribute specifies
the replacement text.

=head2 request

The C<request> element describes a complete HTTP request to be used in the test.
The C<request> has a required attribute, I<url>, which specifies the URL to
which the request is made. This URL may contain a query string. Because of the
requirements of XML, this query string will need to be modified if it contains
the ampersand character E<amp>. The ampersand must be replaced by "E<amp>amp;"
in the I<url> attribute. This string will be converted back to its proper form
before it is requested.

The C<request> element also has an optional I<method> attribute, which defaults
to I<GET>. Currently, B<HTTPtest> supports values of I<GET> and I<POST> for this
attribute. At some point, the program may be extended to support all of the HTTP
request methods.

The C<request> element can contain zero or more C<header> elements followed by
zero or more C<body> elements. In addition, C<include> elements may optionally
occur anywhere in the C<request> content. The C<header> elements allow you to
specify MIME headers to apply to the request. The C<body> element is used to
contain the content of a I<POST> request. At this time, only one C<body> element
is supported in a request. However, at a future date, this will be extended to
multiple C<body> elements for support of the I<multipart> Content types.

=head2 response

The C<response> element defines the subtests to execute for the current test.
After the HTTP request has been made and an HTTP response is returned,
B<HTTPtest> walks the children of the C<response> element and attempts to verify
the response.

The C<response> element has one optional attribute. The I<label> attribute defines
a displayable description of this response set that may be used in error messages
or logging.

The content of the C<response> response element must include one or more of the
following elements:

=over 4

=item *

status

=item *

header

=item *

time

=item *

content

=item *

code

=item *

msg

=item *

filter

=item *

and

=item *

or

=item *

and-all

=item *

or-all

=item *

not

=item *

call

=back

The C<response> element may also contain C<include> elements. After the
C<include> is resolved, the only child elements allowed in a C<response> are
the ones listed above.

=head2 session

The C<session> element is a deprecated synonym for C<test-set>.

=head2 skip

The C<skip> element declares a Perl expression that can determine if the containing
C<test> test should not be executed. This expression can be used to test for
command line options, system resources, current date or time, or anything else
that might cause the running of the test to be a bad idea. The C<skip> element
has a required attribute I<if> that specifies the Perl expression that determines
the need to skip the test. If the expression evaluates to a B<true> value, the
test is skipped, otherrwise the test is performed as normal. The C<skip> element
also supports an optional I<label> attribute.

If the skip test succeeds and HTTPtest is running in verbose mode, the contents
of the C<skip> element are displayed as an explanation for the test being
skipped. If the content is empty, the C<label> is used. If there is no label,
the actual test expression is displayed.

Skipped tests are reported separately from successful or failed tests. This
allows you to count them whichever way makes the most sense.

=head2 status

The C<status> element specifies values to match for the HTTP Response status
code. The element has one required attribute and one optional attribute. The
required I<code> attribute specifies a value to use in the comparison. This
value is interpreted based on the value of the optional I<match> attribute.
If the value of the I<match> attribute is I<exact> or the attribute is missing,
the code must match the response exactly. If the value of the I<match> attribute
is I<regexp>, the value of I<code> is treated as a regular expression.

=head2 subtest_def

The C<subtest_def> element supports naming a group of subtests so that they
may be used from multiple C<response>s.

=head2 test

The C<test> element describes a particular test to be run. The C<test> contains only
zero or more C<skip> elements, a C<request>, and a C<response>. The C<test> also
has one required attribute I<label>, which is a displayable description of the
test which is used in failure reporting.

=head2 testset

The C<testset> element is a deprecated synonym for C<battery>.

=head2 test-set

The C<test-set> element groups a set of C<test> elements together as a single
group.

=head2 time

The C<time> element specifies the expected time limits for an HTTP response. The
C<time> element has two optional attributes, I<min> and I<max>. If the I<min>
attribute is specified, the test fails if the HTTP response returns in less than
that many seconds. If the I<max> attribute is specified, the test fails if the HTTP
response returns in more than that many seconds. If the value of I<min> is greater
than the value of I<max>, the test will always fail.

=head1 Example Tests

The following is a very simple test. We are just testing to see if the page is returned
in a reasonable amount of time. In this case, the test is successful if we recieve a
status code of 200 (OK) and if the page returns in no more than 10 seconds.

  <test label="Did it come back">
    <request url="http://www.tscn.com/" method="GET"/>
	<response>
	  <status code="200"/>
	  <time max="10"/>
	</response>
  </test>

To extend this test a little, we will check to see if the returned page is an HTML document.
In this case, I am taking advantage of the fact that I know the page has the HTML tags in
all upper case.

  <test label="Respond in HTML">
    <request url="http://www.tscn.com/" method="GET"/>
	<response>
	  <status code="200"/>
	  <time max="10"/>
	  <header name="Content-Type" value="text/html"/>
	  <content match="regexp"><![CDATA[<HTML>.*</HTML>]]>
	</response>
  </test>

Now we will check to see if we are getting the page (or, at least the title) that we
think we are getting back.

  <test label="Old WSC root page">
    <request url="http://www.tscn.com/" method="GET"/>
	<response>
	  <status code="200"/>
	  <time max="10"/>
	  <header name="Content-Type" value="text/html"/>
	  <content match="regexp"><![CDATA[<HTML>.*</HTML>]]>
	  <content match="regexp"><![CDATA[
<TITLE>Wall Street City - The Investor's Supersite on the Web</TITLE>
]]>
	</response>
  </test>

=head1 Example msg Tests

The C<msg> element is a little more subtle than most of the other subtests. It
is almost always used in conjunction with the C<or> or C<and> elements. The C<msg>
element serves two purposes: display a message alerting the user to an unusual condition,
abort a battery in case an exceptional situation is detected.

In order to report an exceptional condition that has been recognized in the test, use
the C<and> element. For example,

  <test label="Part of a long sequence of tests">
    <request url="http://www.test.com/" method="GET"/>
	<response>
	  <status code="200"/>
      <and>
        <content match="regexp">This is exceptional</content>
        <msg action="stop" print="always">
            WARNING: Exceptional condition detected.
            Aborting test set...
        </msg>
      </and>
	</response>
  </test>

If, on the other hand, you want to report the failure of a critical subtest on
that would affect any tests that follow, use the C<or> element. For example,

  <test label="First test in a long set">
    <request url="http://www.test.com/" method="GET"/>
	<response>
	  <status code="200"/>
      <or>
        <content match="regexp">This MUST be found</content>
        <msg action="stop" print="always">
            WARNING: Required content missing.
            Aborting test set...
        </msg>
      </or>
	</response>
  </test>

=head1 The Include Path

B<HTTPtest> will search the include path starting with the paths specified in the B<-I>
command line switch, in the order specified. Next, any paths specified in C<option>
elements in the order they are specified in the test set file. Finally, as a last resort,
the subdirectory I<lib> under the directory from which B<HTTPtest> was run is checked.

=head1 Outstanding Issues

The following are issues still outstanding:

=over 4

=item *

The C<codelib> functionality has not been extensively tested.

=item *

The C<code> functionality has not been extensively tested. In addition,
it's functionality may change slightly in the near future.

=item *

Need to clean up the reporting and logging subsystems.

=item *

I do not like the name of the C<inclusion> element. It is subject to change.

=item *

The B<-f suite> command line option which names a test suite file containing the
names of battery files. Need to check handling of params in this case.

=item *

Need to support multiple C<body> tags inside of a C<request>. Also need to check
definition of boundaries and multiple levels of nesting.

=item *

Need to look into other HTTP request methods to verify that the script can support
them and add them to the DTD and input function.

=item *

Once I figure out what a C<test-set> is, I need to implement it. Need to extend
the C<test-set>s so they are more than just a dumb container.

=item *

Need to add functionality similar to C<param> for C<include> elements. This new
child of C<include> would declare replaceable parameters that would be replaced in
the body of the included file. These templatized includes could simplify creation
and maintenance of similar tests.

=item *

I need to add tutorial docs for the use of C<code> and C<codelib>, especially
the new I<id> feature of C<codelib>. 

=back

=cut

