#!/usr/bin/perl
# File: XMLdata.pm
#
#  Released under the same license as Perl.
#  Simplify working with XML::Parser Objects.
#

package XMLdata;
use XMLdata::Element;
use XMLdata::PCData;

use strict;
use vars qw/$xlf $Pkg %convert/;

use constant BaseName => "XMLdata::Element";

$Pkg = 'obj';

#-------------------------------------------------------------------
# For character conversions
%convert = ( '&'=>'&amp;', '<'=>'&lt;', '>'=>'&gt;', '"'=>'&quot;' );

$xlf = make_text( "\n" );  # used for formatting



#--------------------------------------------------------------------
#   Set the current package name and setup the Characters class
#
#   Input:   $pkg  -  package name to use
sub   set_package
 {
  no strict "refs";
  my $pkg = shift;
  $Pkg = $pkg;

  my $name = $Pkg . '::Characters';

  push @{$name.'::ISA'}, "XMLdata::PCData";
 }

#-------------------------------------------------------------------
#  Convert illegal XML characters into appropriate entities.
#
#  Input:  string possibly containing invalid characters
#
#  Return: converted string
sub  xml_convert
 {
  my $str = shift;

  $str =~ s/([&><"])/$convert{$1}/smg;

  $str;
 }



#---------------------------------------------------------------
#  Create an XMLdata object
#
#  Input:  tag name
#		   optional reference to a hash of attribute name/value pairs
#          optional reference to a list of objects to use as kids
#
#  Return: new XMLdata object
sub   create
 {
  my $tag  = shift;
  my $attr = shift || {};
  my $kids = shift || [];

  $tag = $Pkg."::$tag" unless $tag =~ /::/;

  # if content is a string
  $kids = [ make_text( $kids ) ] if '' eq ref $kids;

  bless { %{$attr}, Kids=>$kids }, $tag;
 }




#---------------------------------------------------------------
#  Clone an XMLdata object
#
#  Input:  an XMLdata object
#
#  Return: new XMLdata object that is identical to the supplied
#          object
#  WARNING: This function duplicates an element and all of
#           its children.
sub   clone
 {
  my $obj  = shift;
  my $clone;

  if(is_element( $obj ))
   {
    my $attr = { map { $_ => $obj->{$_} } $obj->get_attribute_names() };
    my $kids = [ map { clone($_) } @{$obj->get_children()} ];

    $clone = create( $obj->tagname(), $attr, $kids );
   }
  elsif(is_text( $obj ))
   {
    $clone = make_text( get_text( $obj ) );
   }
  else
   {
    die "Unrecognized object type '", (ref $obj), "'.\n";
   }

  $clone;
 }




#---------------------------------------------------------------
#  Create an XMLdata character string
#
#  Input:  a character string
#
#  Return: a Characters object containing the text
sub   make_text
 {
  my $str  = shift;

  bless { Text=>$str }, $Pkg."::Characters";
 }




#----------------------------------------------------------------
#  Extract the root element from the reference returned by the
#  Object style of XML::Parser
#
#  Input:   return value from XML::Parser->parse()
#
#  Return:  the xml object representing the root element of the
#           XML document
sub   root
 {
  (grep { (ref $_) !~ /::(Characters|PI|Comment)$/ } @{$_[0]})[0];
 }


#-----------------------------------------------------------------
#  Convert an empty-element object into a string.
#
#  Input:  an xml object
#
#  Return: a scalar containing the xml empty-element as a string
sub   empty_element_to_string
 {
  my $obj = shift;
  my $tag = $obj->tagname();
  
  join( ' ', "<$tag", map { qq{$_="} . xml_convert( $obj->{$_}) . qq{"} } 
             sort @{$obj->get_attribute_names()} ) . "/>";
 }



#-----------------------------------------------------------------
#  Convert an element (including children) into a string. Also
#  handles the array reference returned by XML::Parser in the
#  Object style.
#
#  Input:  an xml object
#
#  Return: a scalar containing the xml element as a string,
#          including it's contents
sub   element_to_string
 {
  my $obj = shift;
  my $tag = ref $obj;

  if('ARRAY' eq $tag)
   {
    my $output = '';
    
    foreach(@{$obj})
     {
      $output .= element_to_string( $_ );
     }

    return $output;
   }
  elsif($obj->is_text())
   {
    return xml_convert( $obj->{Text} );
   }
  elsif($tag =~ /::Comment$/)   # not part of standard Object Style
   {
    return "<!-- $obj->{Text} -->";
   }
  elsif($tag =~ /::PI$/)		# not part of standard Object Style
   {
    return "<?$obj->{Target} $obj->{Text}?>";
   }
  elsif($obj->is_empty())
   {
    return empty_element_to_string( $ obj );
   }

  $tag = $obj->tagname();
  my $output = join( ' ', "<$tag", map { qq{$_="}. xml_convert( $obj->{$_} ). qq{"} } 
                          sort @{$obj->get_attribute_names()} ) . ">";

  foreach my $child ($obj->get_children())
   {
    $output .= element_to_string( $child );
   }

  $output .= "</$tag>";
  $output;
 }



#-----------------------------------------------------------------
#  Execute the supplied function on the supplied object and all of
#  its children.
#
#  Input:  $obj  - an xml object
#          $exec - a code reference to execute on $obj and all of
#                  its descendents
sub   visit_elements
 {
  my $obj  = shift;
  my $exec = shift;

  $exec->( $obj, @_ );

  return unless $obj->is_element();

  foreach($obj->get_children())
   {
    visit_elements( $_, $exec, @_ );
   }
 }



#-----------------------------------------------------------------
#  Remove any ignorable whitespace from the tree of elements
#  supplied.
#
#  Definition: Ignorable whitespace in this context meets the
#              following criteria:
#              1. Must not contain non-whitespace characters
#              2. Is not the only content of an element
#              3. All other text content of this element must
#                 also be ignorable whitespace.
#
sub  remove_ignorable_whitespace
 {
  my $obj = shift;
  my $tag = ref $obj;

  if('ARRAY' eq $tag)
   {
    my @arr = ();
	foreach my $node (@{$obj})
	 {
	  if((ref $node) =~ /::Characters$/)
	   {
	    return if $node->{Text} =~ /\S/;  # leave it if non-space character
		next;
	   }
	  elsif((ref $node) =~ /::/)
	   {
	    remove_ignorable_whitespace( $node );
	   }

	  push @arr, $node;
	 }
	@{$obj} = @arr;
   }
  elsif($tag =~ /::/)
   {
    return if 2 > $obj->count_children();  # empty or single child

    remove_ignorable_whitespace( $obj->{Kids} );
   }
  else
   {
    die "Unrecognized object type \'$tag\'.\n";
   }
 }





#-----------------------------------------------------------------
#  Insert indent before children
#
#  Inputs:  $obj    - reference container object
#           $indent - string or text object to put before each
#                     child
sub   indent_children
 {
  my $obj    = shift;
  my $indent = shift;

  $indent = XMLdata::make_text( $indent ) unless ref $indent;

  @{$obj->{Kids}} = map { $indent, $_ } @{$obj->{Kids}};
 }





#-----------------------------------------------------------------
#  Make package associated with supplied element name a derived
#  class of XMLdata::Element
#
#  Inputs:  $name   - name of an element class to derive from
#                     XMLdata::Element
sub   DeclareElement
 {
  no strict "refs";
  my $name = shift;

  $name = $XMLdata::Pkg.qq{::$name} unless $name =~ /::/; 

  # don't add it again.
  return if grep { BaseName eq $_ } @{$name.'::ISA'};

  push @{$name.'::ISA'}, BaseName;
 }



#-----------------------------------------------------------------
#  Make packages associated with supplied element names derived
#  classes of XMLdata::Element
#
#  Inputs:  @_   - list of names of element classes to derive from
#                  XMLdata::Element
sub   DeclareElements
 {
  foreach (@_)
   {
    DeclareElement( $_ );
   }
 }

1;
