#!/usr/bin/perl
# File: XMLdata/Element.pm
#
#  Released under the same license as Perl.
#  Simplify working with XML::Parser Objects.
#
package XMLdata::Element;

use strict;


#---------------------------------------------------------------
#  Return the tag name from an xml object (w/o the package)
#
#  Return:  tag name as a string
sub   tagname
 {
  my $self = shift;
  my $tag = ref $self;

  $tag =~ /([^:]+)$/;

  $1;
 }


#----------------------------------------------------------------
#  Return a list of attribute names for the object.
#
#  Return: a list (or reference to a list in scalar context) of
#          the names of all attributes of the supplied object
sub   get_attribute_names
 {
  my $self = shift;
  my @attr = grep { 'Kids' ne $_ } keys %{$self};

  wantarray ? @attr : \@attr;
 }



#----------------------------------------------------------------
#  Returns the number of children in the element
#
#  Return: the count of the number of children of the element
sub   count_children
 {
  my $self = shift;

  defined $self->{Kids} ? scalar @{$self->{Kids}} : 0;
 }



#----------------------------------------------------------------
#  Returns true if the object has no children.
#
#  Return: true if the element has no children, false otherwise
sub   is_empty
 {
  my $self = shift;

  (0 == $self->count_children());
 }




#---------------------------------------------------------------
#  Return a list (or a reference to a list) of children for the
#  object.
#
#  Return: a list (or reference to a list in scalar context) of
#          xml objects
sub   get_children
 {
  my $self = shift;

  if(exists $self->{Kids})
   {
    wantarray ? @{$self->{Kids}} : $self->{Kids};
   }
  else
   {
    wantarray ? () : [];
   }
 }


#---------------------------------------------------------------
#  Return a list (or a reference to a list) of children of the
#  given type from the object.
#
#  Input:  $tag  - a tag name
#
#  Return: a list (or reference to a list in scalar context) of
#          xml objects
sub   get_children_by_tag
 {
  my $self = shift;
  my $tag  = shift;

  $tag = $XMLdata::Pkg."::$tag" unless $tag =~ /::/;

  my @children = grep { $tag eq ref $_ } $self->get_children();

  wantarray ? @children : \@children;
 }


#---------------------------------------------------------------
#  Return a list (or a reference to a list) of children matching
#  the given attributes
#
#  Input:  $attr  - an attribute name
#          $value - value of that attribute
#
#  Return: a list (or reference to a list in scalar context) of
#          xml objects
sub   get_children_by_attribute
 {
  my $self  = shift;
  my $attr  = shift;
  my $value = shift;

  return wantarray ? () : [] unless defined $attr;

  my @children;
  
  if(defined $value)
   {
    # attribute value matches
    @children = grep { exists $_->{$attr} and $value eq $_->{$attr} }
                       $self->get_children();
   }
  else
   {
    # attribute exists
    @children = grep { exists $_->{$attr} } $self->get_children();
   }

  wantarray ? @children : \@children;
 }



#---------------------------------------------------------------
#  Return a list (or a reference to a list) of descendents of the
#  given type from the object.
#
#  Input:  an xml object
#          $tag  - a tag name
#
#  Return: a list (or reference to a list in scalar context) of
#          xml objects
sub   get_descendents_by_tag
 {
  my $self = shift;
  my $tag  = shift;

  $tag = $XMLdata::Pkg."::$tag" unless $tag =~ /::/;

  my @descend = ();

  foreach my $child ($self->get_children())
   {
    push @descend, $child if $tag eq ref $child;
    push @descend, $child->get_descendents_by_tag( $tag )
                    unless !$child->isa('XMLdata::Element') or $child->is_empty();
   } 

  wantarray ? @descend : \@descend;
 }


#----------------------------------------------------------------
#  Return a particular child by index of the object.
#
#  Input:  $n    - an optional index
#
#  Return: an xml objects
sub   get_child
 {
  my $self = shift;
  my $n    = shift || 0;

  $self->get_children()->[$n];
 }


#----------------------------------------------------------------
#  Return a particular child of the given type by index of the
#  object.
#
#  Input:  $tag  - a tag name
#          $n    - an optional index
#
#  Return: an xml objects
sub   get_child_by_tag
 {
  my $self = shift;
  my $tag  = shift;
  my $n    = shift || 0;

  $self->get_children_by_tag( $tag )->[$n];
 }


#----------------------------------------------------------------
#  Return a particular child matching the given attributes by
#  index of the supplied object.
#
#  Input:  $attr  - a tag name
#          $value - value of that attribute
#          $n     - an optional index
#
#  Return: an xml objects
sub   get_child_by_attribute
 {
  my $self  = shift;
  my $attr  = shift;
  my $value = shift;
  my $n     = shift || 0;

  $self->get_children_by_attribute( $attr, $value )->[$n];
 }


#----------------------------------------------------------------
#  Add children to the end of the object's content
#
#  Input:  a list of xml objects to append
sub   append_children
 {
  my $self   = shift;

  push @{$self->{Kids}}, @_;
 }




#----------------------------------------------------------------
#  Replace children by index of the object.
#
#  Input:  $ndx  - index of the child in the container
#          $num  - number of items to remove
#                - list of xml objects to insert
#
#  Return: a list of child xml objects replaced
sub   replace_children
 {
  my $self = shift;
  my $ndx  = shift;
  my $num  = shift;

  splice( @{$self->{Kids}}, $ndx, $num, @_ );
 }




#----------------------------------------------------------------
#  Remove children by index of the object.
#
#  Input:  $ndx  - an optional index
#          $num  - number of items to remove
#
#  Return: a list of child xml objects removed
sub   remove_children
 {
  my $self = shift;
  my $ndx  = shift;
  my $num  = shift;

  $self->replace_children( $ndx, $num );
 }





#----------------------------------------------------------------
#  Remove children by index of the object.
#
#  Input:  $ndx  - an optional index
#
#  Return: the child xml object removed
sub   remove_child
 {
  my $self = shift;
  my $ndx = shift;

  splice( @{$self->{Kids}}, $ndx, 1 );
 }




#----------------------------------------------------------------
#  Return the text contained in a given element
#
#  Input:  an xml object with containing text
#
#  Return: the text string from the content
sub   get_text_content
 {
  $_[0]->get_child( 0 )->get_text();
 }



#----------------------------------------------------------------
#  Return a boolean stating that this is an Element
sub   is_element
 {
  1;
 }




#----------------------------------------------------------------
#  Returns true if the supplied object is a text object
#
#  Input:  an xml object
#
#  Return: true if the xml object is a text object, false otherwise
sub   is_text
 {
  0;
 }
1;
